# ==============================================================================
# Copyright [2013] [Kevin Carter]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
__author__ = "Kevin Carter"
__contact__ = "Kevin Carter"
__email__ = "kevin.carter@rackspace.com"
__copyright__ = "2013 All Rights Reserved"
__date__ = "2013-23-03"
__version__ = "0.003"
__status__ = "Seems to Work"
__appname__ = "xenosimager"
__license__ = "Apache2.0"
__description__ = ("Create an image of a instance from data obtained from"
                   " within the instance")
__urlinformation__ = "http://github.com/cloudnull/transporter"

# The Version Of the Application
V_N = '%s' % __version__

# The Version and Information for the application
VINFO = ('%(version)s, developed by %(author)s, '
         'FYI : %(copyright)s' % {'version': __version__,
                                  'author': __author__,
                                  'copyright': __copyright__})




# ==============================================================================
# Copyright [2013] [Kevin Carter]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
prefabs = {
        'image_done':
                {'subject': 'Image is done Baking',
                 'body': ('Good Day,\n\n'
                          '  This message is to inform you that the'
                          ' image name:\n"%(name)s"\n\nis complete and'
                          ' %(status)s. The ID of this image is "%(uuid)s".')},
        'crusty_image':
                {'subject': 'Stale Images Found',
                 'body': ('Howdie from the Cloud,\n\n'
                          '  Durring Imaging we found an old crusty image.'
                          ' The Image ID was "%(id)s" and the last time it was'
                          ' updated was "%(updated)s". This image was nuked'
                          ' and we are attempting to'
                          ' create another one. As such you should be aware'
                          ' that the image was found in a "%(status)s" state'
                          ' and was old. So at some point in recent history'
                          ' things were not working as intended.')},
        'never_active':
                {'subject': 'Image Never Went Active',
                 'body': ('Salutations,\n\n'
                          '  I am sorry to say that the image process has'
                          ' failed. The Image ID was "%(id)s" while the process'
                          ' looked like it would succeed, the image never went'
                          ' active. This failed image was nuked, while this can'
                          ' happen for a lot of reasons I would recommend you'
                          ' have a look at the instance to see if you have a'
                          ' bazillion INODES or if the instance is simply'
                          ' taking up too much space, which would make it less'
                          ' than practicle to create instance images. Otherwise'
                          ' you can login and fire off a manual image or see'
                          ' if this happens again on the next run.')},
        'image_fail':
                {'subject': 'Sorry to Say but We have had Failures',
                 'body': ('Dear User,\n\n'
                          '  Its with regret that I inform you that while'
                          ' attempting to create an image of your instance,'
                          ' we have failed. We are not 100 percent sure why'
                          ' this happened, however here is what we do know.'
                          '\n\nERROR = %(data)s\nTRACE = %(trace)s')}
                }

footer = ('\nBest Regards,\n\n--\n\nThe CloudNull Self Imager.'
          ' By Kevin Carter\nCheck me out at http://cloudnull.io\n'
          'Experience... Fanatical Support!\n')



# ==============================================================================
# Copyright [2013] [Kevin Carter]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import smtplib
import time
from email.mime.text import MIMEText
import traceback

from imager.info import __appname__
from imager import utils, prefabs, get_instance_name


class Mailer(object):
    def __init__(self, notice, args, logger):
        """
        Send a message to the system administrator about and issues that may
        be happening. If Email is detected as a no go, the system will log the
        message.
        """
        self.logger = logger
        _notice = {'time': '%s UTC' % time.strftime("%H:%M").lstrip(),
                   'random_string': utils.rand_string(),
                   'appname': __appname__,
                   'contact': args['mail_main_contact']}

        for _kv in notice.keys():
            _notice[_kv] = notice[_kv]
        self.notice = _notice
        self.msg = prefabs.prefabs[self.notice['reason']]
        self.body = self.msg['body'] % self.notice

        # Set Mailer Tuple, to work all values have to be true
        mailer = (args['mail_url'],
                  args['mail_password'],
                  args['mail_user'],
                  args['mail_port'])
        if all(mailer):
            # Auth details for the Mail Server
            smtppath = '%s' % mailer[0]
            password = '%s' % mailer[1]
            username = '%s' % mailer[2]
            smtpport = '%s' % mailer[3]

            self.smtp = smtplib.SMTP(host=smtppath, port=smtpport)
            if args['mail_tls']:
                if args['mail_key'] and args['mail_cert']:
                    keyfile = args['mail_key']
                    certfile = args['mail_cert']
                    self.smtp.starttls(keyfile, certfile)
                else:
                    self.smtp.starttls()

            if username and password:
                self.smtp.login(user=username, password=password)
            elif username and not password:
                self.smtp.login(user=username, password=None)
            elif not username and password:
                self.smtp.login(user=None, password=password)
            else:
                self.smtp.login(user=None, password=None)

            if args['mail_debug']:
                self.smtp.set_debuglevel(True)

            # Deliver Cus Messages
            self.cus_messages()

            # Stop SMTP
            self.smtp.quit()
        else:
            self.logger.warn('No mail facility has been set. So no notification'
                             ' is being sent. Message would have been ==>\t'
                             ' "%s"' % self.body)

    def cus_messages(self):
        """
        Generate Customer Emails
        """
        try:
            mailbody = self.body
            self.logger.info(mailbody)
            customer_message = mailbody + prefabs.footer
            customer_message = customer_message.encode('utf8')
            hostname = get_instance_name()
            em_msg = MIMEText(customer_message, 'plain', None)
            em_msg["Subject"] = '%s - %s' % (self.msg["subject"], __appname__)
            em_msg["From"] = "%s@%s" % (__appname__, hostname)
            em_msg["To"] = self.notice['contact']
            em_msg["Reply-To"] = 'DoNotReply@%s' % hostname

            self.logger.info('Delivering email to %(contact)s at "%(time)s for'
                             ' %(reason)s"' % self.notice)

            # Send Customer Messages
            self.smtp.sendmail(em_msg["From"],
                               [em_msg["To"]],
                               em_msg.as_string())
        except Exception:
            self.logger.warn('Mail Failed to deliver, Error'
                             ' Providing Messaging Services\n\n%s'
                             % traceback.format_exc())



# ==============================================================================
# Copyright [2013] [Kevin Carter]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import time
import traceback
import sys
from bookofnova import computelib
from imager import utils


class UnknownStatusError(Exception):
    pass


class FiveHundredApiinfor(Exception):
    pass


class NovaFunctionalLogic(object):
    def __init__(self, logger, args):
        """
        Perform NOVA Commands and then make sure I was not dumb about it
        """
        self.logger = logger
        self.args = args

    def nova_errors(self, status, nova=None, action=None, optional=None):
        """
        Check for errors and see if there were issues. Return False if there
        were problems, return True if all is well.
        """
        if 'nova_status' in status and 'nova_reason' in status:
            msg = ('API ERROR ==> %s ==> %s'
                   % (status['nova_status'], status['nova_reason']))
        else:
            msg = ('Shits Broke API is Busted. Here is what I do know ==> %s'
                   % status)

        if status['nova_status'] > 500:
            self.logger.critical(msg)
            return False
        elif status['nova_status'] == 413:
            self.logger.warn(msg)
            if ('retry-after' in status['nova_reason'] and
                int(status['nova_reason']['retry-after']) == 0):
                info = ('We recieved a Nova error 413 and the retry length was'
                        ' 0. Generally this is because the compute node is out'
                        ' of resources, as such. I died. here are some details'
                        ' on the Status Data ==> %s' % status)
                self.logger.error(info)
                return False
            else:
                _time = int(status['nova_reason']['retry-after'])
                time.sleep(_time)
                if nova:
                    nova.re_authenticate()
                return False
        elif status['nova_status'] == 401:
            self.logger.warn(msg)
            if nova:
                nova.re_authenticate()
            return False
        elif status['nova_status'] == 404:
            if action == 'nuker':
                self.logger.info('Image ID "%s" was not found, likely'
                                 'already Gone' % optional)
                return True
            else:
                info = ('API Action URI Not Found Info we have ==> %s' % status)
                self.logger.warn(info)
            return False
        elif status['nova_status'] == 409:
            self.logger.warn(('API ERROR ==> %s ==> Imaging is in process,'
                              ' please wait, or manuallly delete the present'
                              ' image and try again' % status['nova_status']))
            sys.exit()
        elif status['nova_status'] >= 300:
            self.logger.warn(msg)
            return False
        else:
            return True

    def os_auth(self):
        """
        Perform Authentication and Provide a pre-authenticated Nova method
        """
        for retry in utils.retryloop(attempts=2, timeout=120, delay=15):
            try:
                nova = computelib.NovaCommands(m_args=self.args,
                                               output=self.logger)
                auth_dict = nova.auth()
                error_check = self.nova_errors(status=auth_dict,
                                               action='authentication')
                if error_check:
                    return nova
                else:
                    try:
                        retry()
                    except utils.RetryError:
                        self.logger.critical('NOVA Logic Failed to perform'
                                             ' an error check')
                        return False
            except Exception:
                self.logger.warn(traceback.format_exc())
                try:
                    retry()
                except utils.RetryError:
                    self.logger.critical('Failed to Authenticate after'
                                         ' multiple retries')
                    return False

    def list_image(self, nova):
        """
        List all of the instances using details
        """
        for retry in utils.retryloop(attempts=2, timeout=120, delay=15):
            self.logger.info('Listing Images')
            try:
                nova_list = nova.image_list_detail()
                error_check = self.nova_errors(status=nova_list,
                                               action='nova_list')
                if error_check:
                    return nova_list['nova_resp']
                else:
                    try:
                        retry()
                    except utils.RetryError:
                        return False
            except Exception:
                self.logger.critical('Error While Listing Instances')
                try:
                    retry()
                except utils.RetryError:
                    self.logger.critical('Failed Retry loop while listing'
                                         ' an images')
                    return False

    def destroy_image(self, image_id, nova):
        """
        Delete a running instance.
        """
        try:
            for retry in utils.retryloop(attempts=2,
                                              timeout=120,
                                              delay=15):
                self.logger.warn('Destroying Image ID "%s"' % image_id)
                try:
                    nuked = nova.image_nuker(image_id)
                    error_check = self.nova_errors(status=nuked,
                                                   action='nuker',
                                                   optional=image_id)
                    if error_check:
                        return True
                    else:
                        try:
                            retry()
                        except utils.RetryError:
                            self.logger.critical('Failed Retry loop while'
                                                 ' nuking image')
                            return False
                except Exception:
                    error = traceback.format_exc()
                    self.logger.error('Issues while Destroying Image\n\n%s'
                                      % error)
                    try:
                        retry()
                    except utils.RetryError:
                        self.logger.critical('Failed Retry loop while'
                                             ' destoying an image')
                        return False
        except utils.RetryError:
            self.logger.critical('Failed Retry loop while deleting an image')
            return False

    def image_create(self, nova, server_id, img_name, meta_data=None):
        """
        Create an image of a server
        """
        for retry in utils.retryloop(attempts=2, timeout=120, delay=15):
            self.logger.info('Listing Instaces')
            try:
                nova_img = nova.image_create(server_id, img_name, meta_data)
                error_check = self.nova_errors(status=nova_img,
                                               action='nova_list')
                if error_check:
                    return True
                else:
                    try:
                        retry()
                    except utils.RetryError:
                        return False
            except Exception:
                self.logger.critical('Error While Creating Image')
                try:
                    retry()
                except utils.RetryError:
                    self.logger.critical('Failed Retry loop'
                                         ' while creating Image')
                    return False

    def status_active(self, image_id, nova):
        """
        Wait for an image to go active.
        """
        status_counter = 0
        for retry in utils.retryloop(attempts=200, timeout=1800, delay=20):
            self.logger.info('Image "%s" is waiting for an'
                             ' "ACTIVE" state' % image_id)
            try:
                _data = nova.image_info(image_id)
                _info = _data['nova_resp']['image']
                error_check = self.nova_errors(status=_data,
                                               action='status_active',
                                               optional=image_id)
                if not error_check:
                    status_counter += 1
                    if status_counter >= 5:
                        return False
                    else:
                        self.logger.error('we recieved a fatal error when'
                                   ' attempting to poll the Status'
                                   ' of Image "%s" If this does'
                                   ' not change in a few attempts we'
                                   ' will rebuild the failing node,'
                                   ' current FAIL COUTER is "%s"'
                                   % (image_id, status_counter))
                        try:
                            retry()
                        except utils.RetryError:
                            self.logger.critical('There was a problem with'
                                                 ' waiting on the the Image'
                                                 ' to go active')
                            return False
                else:
                    status = _info['status'].upper().encode('utf8')
                    if status == "ERROR":
                        self.logger.error('The Image "%s" has gone into an'
                                   ' "ERROR" during the imaging process'
                                   % image_id)
                        return False
                    elif status != "ACTIVE":
                        try:
                            retry()
                        except utils.RetryError:
                            self.logger.critical('The image NEVER went active,'
                                       ' and we were not able to'
                                       ' accomplish the imaging after'
                                       ' 200 attempts. Here is the'
                                       ' error that we had ERROR\t:'
                                       '%s ==> %s'
                                       % (_data['nova_status'],
                                          _data['nova_reason']))
                            return False
                    else:
                        data = {}
                        data['name'] = _info['name']
                        data['status'] = _info['status']
                        data['uuid'] = _info['id']
                        return data

            except Exception:
                self.logger.critical('When POLL\'ing HOST ==> "%s" we'
                                     ' encountered an error. Will retry again.'
                                     % image_id)
                try:
                    retry()
                except utils.RetryError:
                    self.logger.critical('Failed Retry loop while'
                               ' booting an instance')
                    return False



# ==============================================================================
# Copyright [2013] [Kevin Carter]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import sys
import os
import ConfigParser
import codecs
import stat


class ConfigureationSetup(object):
    def __init__(self, args):
        """
        Parse arguments from a Configuration file.  Note that anything can be
        set as a "Section" in the argument file.
        """
        self.args = args
        self.config_file = args['system_config']
        self.check_perms()

    def check_perms(self):
        """
        Check the permissions of the config file prior to performing any actions
        """
        # If config file is specified, confim proper permissions
        if self.config_file:
            confpath = self.config_file
            if os.path.isfile(os.path.realpath(confpath)):
                mode = oct(stat.S_IMODE(os.stat(confpath).st_mode))
                modes = (mode == '0600', mode == '0400')
                if not any(modes):
                    sys.exit('To use a configuration file the permissions'
                             ' need to be "0600" or "0400"')

    def config_args(self):
        """
        Loop through the configuration file and set all of our values.
        """
        # setup the parser to for safe config parsing with a no value argument
        # Added per - https://github.com/cloudnull/turbolift/issues/2
        if sys.version_info >= (2, 7, 0):
            parser = ConfigParser.SafeConfigParser(allow_no_value=True)
        else:
            parser = ConfigParser.SafeConfigParser()

        # Load the configuration file for parsing
        with codecs.open(self.config_file, 'r', encoding='utf-8') as f:
            parser.readfp(f)

        # Ensure that there is atleast one section in the configuration file
        if len(parser.sections()) < 1:
            sys.exit('No sections were placed into the configuration file as'
                     ' such I have quit.')
        else:
            # Parse all sections for the configuration file
            for section_name in parser.sections():
                for name, value in parser.items(section_name):
                    name = name.encode('utf8')
                    value = value.encode('utf8')
                    self.args.update({name: value})

        return self.args



# ==============================================================================
# Copyright [2013] [Kevin Carter]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import argparse
import sys
import os

# Local Modules
from imager import systemconfig, info


def get_all_the_values():
    """
    Look for flags, these are all of the available options.
    """
    parser = argparse.ArgumentParser(
        formatter_class=lambda
        prog: argparse.HelpFormatter(prog,
                                     max_help_position=31),
        usage='%(prog)s',
        description=info.__description__,
        argument_default=None,
        epilog=info.VINFO)

    # Positional Groups being used
    openstack = parser.add_argument_group(title='OpenStack Arguments',
                                           description=('Optional Arguments'
                                                        ' for use with'
                                                        ' OpenStack. NOTE:'
                                                        ' ALL of the Openstack'
                                                        ' Credentials need to'
                                                        ' go into the system-'
                                                        'config file.'))

    mail = parser.add_argument_group(title='Mail Credentials',
                                           description=('Used for external '
                                                        'Mail notifications'
                                                        ))

    parser.add_argument('--system-config',
                        type=str,
                        required=True,
                        metavar='[Path]',
                        default=None,
                        help=('Uses a System Config File, for all'
                              ' checks, any available MySQL DBs, and'
                              ' OpenStack Credentials'))

    parser.add_argument('--log-level',
                        default='info',
                        choices=['debug',
                                 'info',
                                 'warn',
                                 'error',
                                 'critical'],
                        help='Set Log Verbosity')

    parser.add_argument('--image-name',
                        required=True,
                        type=str,
                        help='Set the name of the image')

    mail.add_argument('--mail-password',
                      type=str,
                      metavar='[Mail_Pasword]',
                      default=None,
                      help='Password for the Mail Server')

    mail.add_argument('--mail-user',
                      type=str,
                      metavar='[MailUser]',
                      default=None,
                      help='Username for the mail Account')

    mail.add_argument('--mail-url',
                      type=str,
                      metavar='[MailURL]',
                      default=None,
                      help='URL to the Mail Server')

    mail.add_argument('--mail-port',
                      type=str,
                      metavar='[MailPort]',
                      default=None,
                      help='Por the Mail Server will use')

    mail.add_argument('--mail-tls',
                      action='store_true',
                      default=None,
                      help='Does the Mail Server use TLS')

    mail.add_argument('--mail-key',
                      type=str,
                      metavar='[HostIP]',
                      default=None,
                      help='If using TLS is their a Key File')

    mail.add_argument('--mail-cert',
                      type=str,
                      metavar='[HostIP]',
                      default=None,
                      help='If Using TLS is their a Cert File')

    mail.add_argument('--mail-main-contact',
                      metavar='[Email@Address]',
                      default=None,
                      help='Who should I notify')

    mail.add_argument('--mail-debug',
                      action='store_true',
                      default=None,
                      help='Makes the SMTP Service Noisy')

    openstack.add_argument('--os-version',
                           metavar='[VERSION_NUM]',
                           default='v2.0',
                           help='env[OS_VERSION]')

    openstack.add_argument('--os-verbose',
                            action='store_true',
                            default=None,
                            help=('Make the OpenStack Authentication'
                                  ' Verbose'))

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit('Give me something to do and I will do it')

    # Parse the Arguments that have been provided
    args = parser.parse_args()

    # Change Arguments in to a Dictionary
    args = vars(args)

    # Parse Config File
    if args['system_config']:
        args = (systemconfig.ConfigureationSetup(args).config_args())

    # Check to see if the user Running the application is allowed to do so.
    user = os.getuid()
    if user != 0:
        sys.exit('This program requires root privileges."')

    # Return My Arguments and begin Monitoring
    return args



#!/usr/bin/env python
# ==============================================================================
# Copyright [2013] [Kevin Carter]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import sys
import traceback

# Local Modules
from imager import arguments, imaging, logger


def run_imager():
    """
    Basic Method for starting the application from the launcher
    """
    # Set the Parsed Args
    args = arguments.get_all_the_values()
    try:
        log = logger.load_in(log_level=args['log_level'])
        imaging.InstanceImager(log=log, p_args=args)
    except KeyboardInterrupt:
        sys.exit('AHH FAIL! You Killed me!')
    except Exception:
        print(traceback.format_exc())
        sys.exit('Major FAILURE, please check logs')


if __name__ == '__main__':
    # Run main if the application is executed from the CLI without the launcher
    run_imager()



# ==============================================================================
# Copyright [2013] [Kevin Carter]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import string
import time
import random


# ACTIVE STATE retry loop
# http://code.activestate.com/recipes/578163-retry-loop/
class RetryError(Exception):
    pass


def retryloop(attempts, timeout=None, delay=None, backoff=1):
    """
    Enter the amount of retries you want to perform.
    The timeout allows the application to quit on "X".
    delay allows the loop to wait on fail. Useful for making REST calls.

    Example:
        Function for retring an action.
        for retry in retryloop(attempts=10, timeout=30, delay=1, backoff=1):
            something
            if somecondition:
                retry()
    """
    starttime = time.time()
    success = set()
    for _ in range(attempts):
        success.add(True)
        yield success.clear
        if success:
            return
        duration = time.time() - starttime
        if timeout is not None and duration > timeout:
            break
        if delay:
            time.sleep(delay)
            delay = delay * backoff
    raise RetryError


def rand_string(length=9, chr_set=string.ascii_uppercase):
    """
    Generate a Random string
    """
    output = ''
    for _ in range(length):
        output += random.choice(chr_set)
    return output




# ==============================================================================
# Copyright [2013] [Kevin Carter]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
configfile = """
# ================================= NOTES ======================================
# Use this configuration file to store sensitive information
# The config file REQUIRES at least ONE section for functionality
# Section names don't matter, however they are nice for oganizing data
# This is a sample file not all variables are needed. Use what you want too.
# The basic Healing system requires a database to store checks and instance
# information.

# Note: Not all variables are needed, simply use what you need to. If you give
# the system mail relay information you will get notices when images are
# created, or issues happen. DO NOT USE "os_password" and "os_apikey" together,
# choose one or the other.

# Available System variables :
# ------------------------------------------------------------------------------
# log_level = [info,error,warn,debug]  +  mail-main-contact = WhoReadsEmail
# os_user = YOU                        +  mail_debug = True | False
# os_apikey = SuperSecretInformation   +  mail_url = AURL
# os_auth_url = Authentication URL     +  mail_user = Username
# os_rax_auth = A-LOCATION             +  mail_password = Password
# os_verbose = True | False            +  mail_cert = /location/cert
# os_password = SuperSecretInformation +  mail_key = KeyForCert
# os_tenant = UsernameGenerally        +  mail_tls = True | False
# os_version = v2.0                    +  mail_port = Port Number
# os_region = %(region)s
# ================================= NOTES ======================================

[basic]
log_level = info

[openstack]
os_user = YOU
os_apikey = RANDOM-NUMBERS-AND-THINGS
os_rax_auth = %(region)s

#[Mail]
#mail_user =
#mail_password =
#mail_cert =
#mail_key =
#mail_tls =
#mail_port =
#mail_url =
"""



#!/usr/bin/env python
# ==============================================================================
# Copyright [2013] [Kevin Carter]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import setuptools
import sys
import os
import tarfile
import time
from subprocess import Popen, PIPE

# Local Packages
from imager import strings, info


# Check the version of Python that we have been told to use
if sys.version_info < (2, 6, 0):
    sys.stderr.write('The Ameba System Presently requires'
                     ' Python 2.6.0 or greater\n')
    sys.exit('\nUpgrade python because your version of it is VERY deprecated\n')

with open('README') as r_file:
    long_description = r_file.read()

setuptools.setup(
    name=info.__appname__,
    version=info.__version__,
    author=info.__author__,
    author_email=info.__email__,
    description=info.__description__,
    install_requires=['bookofnova', 'python-dateutil==1.3.0', 'argparse'],
    packages=['imager'],
    long_description=long_description,
    license=info.__license__,
    url=info.__urlinformation__,
    classifiers=[
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Utilities',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ],
    entry_points={
        "console_scripts":
            ["%s = imager.executable:run_imager" % info.__appname__]
    }
)


def config_files_setup(conf=strings.configfile):
    print('Moving the the System Config file in place')
    # setup will copy the config file in place.
    name = 'config.cfg'
    path = '/etc/%s' % info.__appname__
    full = '%s%s%s' % (path, os.sep, name)
    if not os.path.isdir(path):
        os.mkdir(path)
        with open(full, 'w+') as conf_f:
            conf_f.write(strings.configfile)
    else:
        if not os.path.isfile(full):
            with open(full, 'w+') as conf_f:
                conf_f.write(strings.configfile)
        else:
            print('Their was a configuration file found, I am compressing the '
                  'old version and will place the new one on the system.')
            not_time = time.time()
            backupfile = '%s.%s.backup.tgz' % (full, not_time)
            tar = tarfile.open(backupfile, 'w:gz')
            tar.add(full)
            tar.close()
            xen_s = commander(cmd='which xenstore')
            dom_id = commander(cmd='%s read domid' % xen_s)
            region = commander(cmd=('%s read /local/domain/%s/vm-data'
                                    '/provider_data/region'
                                    % (xen_s, dom_id)))
            if region:
                config_f = conf % {'region': region.upper()}
            else:
                config_f = conf % {'region': 'WhereIsThisInstance'}

            with open(full, 'w+') as conf_f:
                conf_f.write(config_f)
    if os.path.isfile(full):
        os.chmod(full, 0600)
    print('Configuration file is ready.  Please set your credentials in : %s'
          % full)


def commander(cmd):
    """
    Run sub-process (SHELL) commands.
    """
    output = Popen(cmd.split(), stdout=PIPE, stderr=PIPE)
    stdout, stderr = output.communicate()
    stdout = stdout.strip()
    if stderr:
        stdout = None
    return stdout


# Run addtional Setup
config_files_setup()



# ==============================================================================
# Copyright [2013] [Kevin Carter]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import sys
import traceback
from subprocess import Popen, PIPE

from imager import novafunctions, utils, notifier, get_instance_name


class ImagingInProcess(Exception):
    pass


class GeneralImagerFailure(Exception):
    pass


class InstanceImager(object):
    def __init__(self, log, p_args):
        """
        Access Nova and create an image of the instance.
        """
        self.m_args = p_args
        self.log = log
        self._nova = novafunctions.NovaFunctionalLogic(logger=log, args=p_args)
        self.nova = self._nova.os_auth()
        self.uuid = self.get_instance_id()
        self.name = get_instance_name()
        tag = 'CloudNull_imageizer'
        self.image_name = '%s_%s_%s_%s' % (self.name,
                                           self.m_args['image_name'],
                                           self.uuid,
                                           tag)
        self.check_for_image()
        self.create_image()

    def commander(self, cmd):
        """
        Run sub-process (SHELL) commands.
        """
        output = Popen(cmd.split(), stdout=PIPE, stderr=PIPE)
        stdout, stderr = output.communicate()
        if stderr:
            raise GeneralImagerFailure(stderr)
        return stdout.strip()

    def get_instance_id(self):
        """
        Get the UUID of the instance that we want to make an image of
        """
        # Check to see if this application will be run as a Daemon
        try:
            xen_s = self.commander(cmd='which xenstore')
            dom_id = self.commander(cmd='%s read domid' % xen_s)
            slice_id = self.commander(cmd='%s read /local/domain/%s/name'
                                      % (xen_s, dom_id))
            slice_id = slice_id.split('-', 1)[1]
            return slice_id
        except Exception:
            _tb = traceback.format_exc()
            self.log.error(_tb)
            sys.exit('%s\n\nSub Process FAILURE!!!' % _tb)

    def check_for_image(self):
        from datetime import datetime as dt, timedelta
        from dateutil import parser
        try:
            img = self._nova.list_image(nova=self.nova)
            for obj in img['images']:
                if (obj['name'] == self.image_name and
                    obj['status'].upper() == 'ACTIVE'):
                    self._nova.destroy_image(image_id=obj['id'],
                                             nova=self.nova)
                elif (obj['name'] == self.image_name and
                    obj['status'].upper() == 'SAVING'):
                    time = obj['updated']
                    _tm = time.replace('Z', '')
                    _time = parser.parse(_tm)
                    _now = dt.utcnow()
                    if (_now - _time) > timedelta(hours=5):
                        obj['reason'] = 'crusty_image'
                        self.notify_me(obj)
                        self._nova.destroy_image(image_id=obj['id'],
                                                 nova=self.nova)
                    else:
                        raise ImagingInProcess('Image ID "%s"' % obj['id'])
        except ImagingInProcess, exp:
            self.log.info('Imaging is presently in progress. %s' % exp)

    def create_image(self):
        try:
            randkey = utils.rand_string()
            self._nova.image_create(nova=self.nova,
                                    server_id=self.uuid,
                                    img_name=self.image_name,
                                    meta_data={'cloudnull': randkey})
            img = self._nova.list_image(nova=self.nova)
            for obj in img['images']:
                if (obj['name'] == self.image_name and
                    obj['status'].upper() == 'SAVING' and
                    'cloudnull' in obj['metadata'] and
                    obj['metadata']['cloudnull'] == randkey):
                    imid = obj['id']
            if imid:
                data = self._nova.status_active(image_id=imid, nova=self.nova)
                if data:
                    data['reason'] = 'image_done'
                    self.notify_me(data)
                else:
                    self._nova.destroy_image(image_id=imid,
                                             nova=self.nova)
                    data['reason'] = 'never_active'
                    self.notify_me(data)
                    raise GeneralImagerFailure('We Failed waiting for image %s'
                                               ' to go active' % imid)
        except Exception, exp:
            data = {'reason': 'image_fail',
                    'data': exp,
                    'trace': traceback.format_exc()}
            self.notify_me(data)

    def notify_me(self, notice):
        notifier.Mailer(notice=notice,
                        args=self.m_args,
                        logger=self.log)
        del notice



# ==============================================================================
# Copyright [2013] [Kevin Carter]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
import os
from imager.info import __appname__


class NoLogLevelSet(Exception):
    pass


class Logging(object):
    def __init__(self, log_level, log_file=None):
        self.log_level = log_level
        self.log_file = log_file

    def logger_setup(self):
        """
        Setup logging for your application
        """
        logger = logging.getLogger("%s" % (__appname__.upper()))

        avail_level = {'DEBUG': logging.DEBUG,
                       'INFO': logging.INFO,
                       'CRITICAL': logging.CRITICAL,
                       'WARN': logging.WARN,
                       'ERROR': logging.ERROR}

        _log_level = self.log_level.upper()
        if _log_level in avail_level:
            lvl = avail_level[_log_level]
            logger.setLevel(lvl)
            formatter = logging.Formatter('%(asctime)s - %(name)s:%(levelname)s'
                                          ' ==> %(message)s')
        else:
            raise NoLogLevelSet('I died because you did not set a known'
                                ' log level')

        # Building Handeler
        if self.log_file:
            handler = logging.FileHandler(self.log_file)
        else:
            handler = logging.StreamHandler()

        handler.setLevel(lvl)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        return logger


def return_logfile(filename):
    """
    Return a path for logging file.

    IF "/var/log/" does not exist, or you dont have write permissions to
    "/var/log/" the log file will be in your working directory
    Check for ROOT user if not log to working directory
    """
    if os.path.isfile(filename):
        return filename
    else:
        user = os.getuid()
        logname = ('%s' % filename)
        if not user == 0:
            logfile = logname
        else:
            if os.path.isdir('/var/log'):
                log_loc = '/var/log'
                logfile = '%s/%s' % (log_loc, logname)
            else:
                try:
                    os.mkdir('%s' % log_loc)
                    logfile = '%s/%s' % (log_loc, logname)
                except Exception:
                    logfile = '%s' % logname
        return logfile


def load_in(log_level='info'):
    """
    Load in the log handler. If output is not None, systen will use the default
    Log facility.
    """
    _file = '%s.log' % __appname__
    _log_file = return_logfile(filename=_file)
    log = Logging(log_level=log_level, log_file=_log_file)
    output = log.logger_setup()
    return output
